"""curlbomb is an HTTP server for serving one-time-use shell scripts

You know all those docs for cool dev tools that start out by telling
you to install their software in one line, like this?

    bash <(curl -s http://example.com/install.sh)

I call that a curl bomb... I don't know if anyone else does.

This script is an HTTP server that will serve that script to a client
exactly once and then quit. Yea, you could just use "python -m http.server", 
really this is just a bit more than that.

MIT Licensed, see LICENSE.txt

Ryan McGuire <ryan@enigmacurry.com>
http://github.com/EnigmaCurry/curlbomb
"""

from .main import *
from . import argparser, settings, server
from . import run, get, put, ping

__version__ = argparser.get_version()



import logging
import glob
from io import BytesIO
import os

from . import run

log = logging.getLogger('curlbomb.ping')

script = """#!/bin/bash
set -e
SSH_PUBKEY="{ssh_id}"
AUTH_FILE="$HOME/.ssh/authorized_keys"

# If ssh dir does not exist, create it:
if [ ! -d "$HOME/.ssh" ]; then
    mkdir $HOME/.ssh
    chmod 711 $HOME/.ssh
fi

# If authorized_keys file does not exist, create it:
if [ ! -f "$AUTH_FILE" ]; then
    touch $AUTH_FILE
    chmod 700 $AUTH_FILE
fi

# Don't add the key if it's already there:
if ! grep -q "$SSH_PUBKEY" $AUTH_FILE; then
    # Detect if the file is blank:
    if [[ $(cat $AUTH_FILE | wc -c) -gt 1 ]]; then
        # Detect if file does not end in a newline, adds one if necessary:
        if [[ $(tail -n 1 "$AUTH_FILE" | wc --lines) -lt 1 ]]; then 
            echo "" >> $AUTH_FILE
        fi
    fi
    echo "$SSH_PUBKEY" >> $AUTH_FILE
fi
"""

class SSHIdentityException(Exception):
    pass

def add_parser(subparsers):
    parser = subparsers.add_parser(
        'ssh-copy-id', help="copy ssh public key to remote authorized_keys file")
    parser.add_argument('identity', metavar="IDENTITY",
                        help="Read identity from SSH identity file (eg ~/.ssh/id_rsa.pub)")
    parser.set_defaults(prepare_command=prepare)

def get_pubkey(identity):
    if not os.path.isfile(identity):
        raise SSHIdentityException("identity file does not exist: {}".format(identity))
    if not identity.endswith(".pub"):
        raise SSHIdentityException("identity file name must end with .pub")

    with open(identity) as f:
        pubkey = f.read().strip()
    assert len(pubkey.splitlines()) == 1, "SSH identity file has more than one line"
    return pubkey
    
def prepare(args, settings, parser):
    args.resource = settings['resource'] = BytesIO(bytes(script.format(ssh_id = get_pubkey(args.identity)), "utf-8"))

    # Call the run prepare method
    args.command = None
    run.prepare(args, settings, parser)



import sys
import re
from io import BytesIO
import logging
import hashlib

import requests

from .gpg import decrypt_resource_if_necessary, verify_resource

log = logging.getLogger('curlbomb.run')

def add_parser(subparsers):
    run_parser = subparsers.add_parser('run', help="Run a local script on the client")
    run_parser.add_argument('-c', '--command', metavar="COMMAND",
                            help="The the shell command to curlbomb into "
                            "(default is to detect #!interpreter ie. the shebang)")
    run_parser.add_argument('--hash', metavar="SHA256", dest="script_hash",
                            help="The sha256 hash of the resource to verify before serving")
    run_parser.add_argument("--signature", metavar=("FILE_OR_URL","GPG_ID"), nargs="+",
                            help="Verify SCRIPT against this GPG signature. "
                            "Can be a file or a http(s) URL. Specify list of GPG IDs to allow, "
                            "otherwise any in your keyring will work.")
    run_parser.add_argument('resource', metavar="SCRIPT", help="path or URL to script, or - to read from STDIN",
                            nargs='?', default=sys.stdin)
    run_parser.set_defaults(prepare_command=prepare)


def prepare(args, settings, parser):
    settings['time_command'] = True
    settings['shell_command'] = args.command

    if args.resource == sys.stdin:
        args.resource = settings['stdin']
    
    if args.resource == settings['stdin'] and settings['stdin'].isatty():
        parser.print_help()
        sys.stderr.write("\nYou must specify a file or pipe one to this command's stdin\n")
        sys.exit(1)
    if settings.get('resource', None) is None:
        if args.resource == settings['stdin'] or args.resource == '-':
            settings['resource'] = BytesIO(settings['stdin'].buffer.read())
        else:
            if re.match("http[s]?://", args.resource): 
                # Download resource from URL:
                log.info("Downloading resource: {}".format(args.resource))
                r = requests.get(args.resource)
                settings['resource'] = BytesIO(r.content)
            else:
                # Read resource from disk:
                settings['resource'] = open(args.resource, 'br')
            
    settings['resource'] = decrypt_resource_if_necessary(settings['resource'])

    if args.script_hash is not None:
        hash = hashlib.sha256(settings['resource'].read()).hexdigest()
        if args.script_hash != hash:
            log.error("Bad resource hash: {}".format(hash))
            sys.exit(1)

    if args.signature is not None:
        if not verify_resource(settings['resource'], args.signature[0], args.signature[1:]):
            log.error("Invalid signature for resource")
            sys.exit(1)

    #Detect if the input has a shebang so we can detect the shell command to display    
    if args.command is None:
        line = settings['resource'].readline(500)
        settings['resource'].seek(0)
        if line.startswith(b'#!'):
            settings['shell_command'] = line[2:].decode("utf-8").rstrip()
        else:
            settings['shell_command'] = "bash"
        



from .server import run_server
from .settings import get_settings
import logging

def main():
    logging.basicConfig(level=logging.WARN)
    
    settings = get_settings()
    return run_server(settings)
    
if __name__ == "__main__":
    exit(main())



import os
import subprocess
from io import BytesIO
import logging

log = logging.getLogger('curlbomb.get')

def add_parser(subparsers):
    get_parser = subparsers.add_parser(
        'get', help='Copy remote files or directories to the server')
    get_parser.add_argument('source', metavar="SOURCE", nargs=1,
                            help="Remote path to copy (or put glob in quotes)")
    get_parser.add_argument('dest', metavar="DEST", nargs='?',
                            help="Local directory to copy to")
    get_parser.add_argument('--exclude', metavar="PATTERN", action='append',
                            help="Exclude files matching PATTERN, "
                            "a glob(3)-style wildcard pattern", default=[])
    get_parser.set_defaults(prepare_command=prepare)


def prepare(args, settings, parser):
    settings['client_quiet'] = True
    settings['receive_postbacks'] = True
    if args.dest is None:
        dest = os.curdir
    else:
        dest = args.dest
    p = subprocess.Popen(['tar','xzv','-C',dest], stdin=subprocess.PIPE)
    settings['log_process'] = p
    settings['log_file'] = p.stdin
    parent_path, path = os.path.split(args.source[0])
    if len(parent_path) == 0:
        parent_path = os.curdir
    exclude_args = " ".join(["--exclude='{}'".format(p) for p in args.exclude])
    args.resource = settings['resource'] = BytesIO(
        bytes('tar czh {exclude} -C "{parent_path}" "{path}"'.format(
            parent_path=parent_path, path=path, exclude=exclude_args), "utf-8"))



import sys
import logging
import socket
import base64
import random
import shlex

from . import run
from . import tls
from . import gpg

log = logging.getLogger('curlbomb.settings')

from .argparser import argparser

def random_passphrase(length):
    return base64.b64encode(bytes(random.sample(range(256), length)),
                            altchars=b'_.').decode("utf-8")[:length]

def get_curlbomb_command(settings, unwrapped=None):
    """Get the curlbomb command

    Inspects settings['unwrapped'] and returns the full curlbomb
    command if True. A wrapper script is returned in the default case.
    
    Alternatively, you can pass unwrapped=True to force getting the
    unwrapped script.
    """
    # Pin tls cert to client command if we are using curl, --ssl, and --pin
    pin_settings=""
    if settings['http_fetcher'].startswith("curl") and \
       settings['ssl'] is not False and settings['pin'] is True:
        pin_settings = " -k --pinnedpubkey 'sha256//{}'".format(settings['ssl_hash'])

    if (settings['unwrapped'] and unwrapped is not False) or unwrapped is True:
        # Get the full unwrapped command:
        knock = ""
        if settings['require_knock']:
            if settings.get('require_knock_from_environment', False):
                # Don't output the actual knock code, but the shell variable name:
                k = "$KNOCK"
            else:
                k = settings['knock']
            if settings['http_fetcher'].startswith("wget"):
                knock = ' --header="X-knock: {}"'.format(k)
            else:
                knock = ' -H "X-knock: {}"'.format(k)

        hostname_header = ""
        if settings['require_hostname_header']:
            if settings['http_fetcher'].startswith("wget"):
                hostname_header = ' --header="X-hostname: $(hostname)"'
            else:
                hostname_header = ' -H "X-hostname: $(hostname)"'

        url = "http{ssl}://{host}:{port}/r{knock}{hostname_header}".format(
            ssl="s" if settings['ssl'] is not False else "",
            host=settings['display_host'],
            port=settings['display_port'],
            knock=knock,
            hostname_header=hostname_header
        )
        stream_url = "http{ssl}://{host}:{port}/s{knock}{hostname_header}".format(
            ssl="s" if settings['ssl'] is not False else "",
            host=settings['display_host'],
            port=settings['display_port'],
            knock=knock,
            hostname_header=hostname_header
        )

        decrypter = ""
        if settings['client_decrypt']:
            decrypter = " | gpg -d"

        logger = ""
        if settings['client_logging']:
            logger = " | tee curlbomb.log"

        if settings['receive_postbacks']:
            callback_cmd="curl{pin_settings} -T - {stream_url}"
            if settings['client_quiet']:
                callback_cmd = " | " + callback_cmd
            else:
                callback_cmd = " | tee >({cmd})".format(cmd=callback_cmd)
            if settings['wget']:
                callback_cmd = (
                    ' && wget -q -O - --post-data="wget post-back finished. '
                    'wget can\'t stream the client output like curl can though '
                    ':(\r\n" {stream_url}')
            logger += callback_cmd.format(
                stream_url=stream_url,
                pin_settings=pin_settings
                )
        
        if settings['shell_command'] is None or settings['survey']:
            cmd = "{http_fetcher}{pin_settings} {url}{decrypter}{logger}".\
                  format(
                      http_fetcher=settings['http_fetcher'],
                      url=url,
                      pin_settings=pin_settings,
                      decrypter=decrypter,
                      logger=logger)
        else:
            cmd = "{shell_command} <({http_fetcher}{pin_settings} {url}{decrypter}){logger}"
            if settings['pipe_to_shell_command']:
                # If we don't care about receiving client output, pipe
                # it rather than redirecting. This is needed for
                # interactive scripts:
                cmd = "{http_fetcher}{pin_settings} {url} | {shell_command}{logger}"
            cmd = cmd.format(
                shell_command=settings['shell_command'],
                http_fetcher=settings['http_fetcher'],
                pin_settings=pin_settings,
                url=url,
                decrypter=decrypter,
                logger=logger)

        return cmd
    else:
        # Get the wrapped version:
        if settings['survey']:
            if settings['require_knock']:
                if settings['http_fetcher'].startswith("wget"):
                    knock_header = ' --header="X-knock: {}"'.format(settings['knock'])
                else:
                    knock_header = ' -H "X-knock: {}"'.format(settings['knock'])
            else:
                knock_header=''
            cmd = "{http_fetcher}{pin_settings} http{ssl}://{host}:{port}/r" + knock_header
        else:
            cmd = "{knock}bash <({http_fetcher}{pin_settings} http{ssl}://{host}:{port})"

        return cmd.format(
            http_fetcher=settings['http_fetcher'],
            ssl="s" if settings['ssl'] is not False else "",
            host=settings['display_host'],
            port=settings['display_port'],
            pin_settings=pin_settings,
            knock="KNOCK={} ".format(
                shlex.quote(settings['knock'])) if settings['require_knock'] else ''
        )

    
def get_settings(args=None, override_defaults={}):
    """Parse args and set other settings based on them

    Default settings can be overriden by passing in a dictionary
    
    Return a new dictionary containing all args and settings
    """
    # Parse the arguments:
    parser = argparser()
    args = parser.parse_args(args)
        
    settings = {
        # Store args object:
        'args': args,
        # Instruct client to post stdout back to the server:
        'receive_postbacks': args.log_post_backs,
        # Run client script with this shell interpreter:
        'shell_command': 'bash',
        # Client fetches URL resources with this command:
        'http_fetcher': 'curl -LSs',
        # Mime type to serve resource as:
        'mime_type': args.mime_type,
        # Client should send it's hostname in the request header:
        'require_hostname_header': False,
        # Log client stdout to server stdout:
        'log_post_backs': args.log_post_backs,
        # Enable TLS - Path to cert or None specifies to generate a self-signed cert
        'ssl': None if args.ssl == "-" else args.ssl,
        # ssl context passed to server
        'ssl_context': None,
        # Enable SSL certificate pinning in client command:
        'pin': args.pin or args.ssl is None,
        # Total number of allowed HTTP gets on resource:
        'num_gets': args.num_gets,
        # Require X-knock header:
        'require_knock': not args.disable_knock,
        # The current knock:
        'knock': args.knock,
        # Server verbose flag
        'verbose': args.verbose,
        # Print curl command without shell_command
        'survey': args.survey,
        # SSH tunnel
        'ssh': args.ssh,
        # Server quiet flag
        'quiet': args.quiet and not args.verbose,
        # Log stdout on client:
        'client_logging': args.client_logging,
        # Client quiet flag
        'client_quiet': args.client_quiet,
        # Client decryption:
        'client_decrypt': args.encrypt or args.passphrase or args.encrypt_to,
        # Server encryption (random_passphrase, user_passphrase, recipient):
        'encryption_type': "user_passphrase" if args.passphrase else "recipient" if args.encrypt_to else "random_passphrase" if args.encrypt else None,
        # the passphrase that was generated or supplied by the user:
        'passphrase': None,
        # Popen object processing log_post_backs
        'log_process': None,
        # File to receive log_post_backs:
        'log_file': None,
        # Don't print knock in wrapped curlbomb command:
        'require_knock_from_environment': True,
        # Client should use wget instead of curl
        'wget': args.wget,
        # Don't wrap curlbomb 
        'unwrapped': args.unwrapped,
        # Use alternative stdin, only used in tests
        'stdin': sys.stdin,
        # Use alternative stdout, only used in tests
        'stdout': sys.stdout,
        # Output how long the command takes:
        'time_command': False,
        # Function to get curlbomb command given settings:
        'get_curlbomb_command': get_curlbomb_command,
        # Normally we do process substituition, but we can do a regular pipe as well:
        'pipe_to_shell_command': args.pipe
    }
    settings.update(override_defaults)
    
    if args.verbose:
        logging.getLogger('curlbomb').setLevel(level=logging.INFO)
        logging.getLogger('tornado.access').setLevel(level=logging.INFO)

    if args.debug:
        settings['verbose'] = True
        logging.getLogger('curlbomb').setLevel(level=logging.DEBUG)
        logging.getLogger('tornado.access').setLevel(level=logging.DEBUG)
        
    if settings['require_knock'] and not settings['knock']:
        settings['knock'] = random_passphrase(12)
        
    if settings['survey']:
        # Don't recieve post backs in survey mode:
        settings['receive_postbacks'] = False
        settings['client_logging'] = False

    if args.disable_postback:
        log.error("--disable-postback and -1 are no longer valid options. If you were using it for interactive script support, use --pipe.")
        sys.exit(1)
        
    if settings['unwrapped']:
        # Output the unrwapped version of the curlbomb Without this
        # setting, curlbomb usually outputs a url that retrieves a
        # wrapper script that wraps the longer more complicated client
        # command. This will output this unrwapped version instead.
        settings['require_knock_from_environment'] = False

    if args.pin and args.ssl is False:
        print("--pin requires --ssl")
        sys.exit(1)

    if settings['encryption_type'] == 'user_passphrase':
        if args.encrypt_to:
            print("--passphrase and --encrypt-to options are incompatible with each other.")
            sys.exit(1)
        if not settings['passphrase']:
            settings['passphrase'] = input("Enter passphrase to encrypt resource: ")
        if len(settings['passphrase']) == 0:
            log.error("Passphrase cannot be blank")
            sys.exit(1)
    elif settings['encryption_type'] == 'random_passphrase':
        settings['passphrase'] = random_passphrase(25)

    if args.wget:
        settings['http_fetcher'] = "wget -q -O -"
        if args.log_post_backs:
            print("wget can't stream the client output, so --log-posts is not "
                  "supported in wget mode")
            sys.exit(1)
        if args.pin:
            print("SSL certificate pinning only works in curl mode")
            sys.exit(1)

    if args.port == "random":
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(('',0))
        settings['port'] = s.getsockname()[1]
        s.close()
    else:
        settings['port'] = int(args.port)

    settings['display_host'] = socket.gethostbyname(socket.gethostname())
    settings['display_port'] = settings['port']

    if settings['ssh']:
        ssh_parts = settings['ssh'].split(":")
        ssh_host = ssh_parts[0]
        ssh_port = 22
        http_port = settings['port']
        if len(ssh_parts) == 3:
            ssh_port = ssh_parts[1]
            http_port = ssh_parts[2]
        elif len(ssh_parts) == 2:
            http_port = ssh_parts[1]
        settings['ssh_forward'] = "0.0.0.0:{http_port}:localhost:{port}".format(
            port=settings['port'], http_port=http_port)
        settings['ssh_host'] = ssh_host
        settings['ssh_port'] = ssh_port
        settings['display_port'] = http_port
        if '@' in ssh_host:
            settings['ssh_user'], settings['display_host'] = ssh_host.split('@')
        else:
            settings['display_host'] = ssh_host

    if args.domain:
        # Override displayed host:port
        parts = args.domain.split(":")
        settings['display_host'] = parts[0]
        if len(parts) > 1:
            settings['display_port'] = parts[1]

    if settings['ssl'] is not False:
        settings['ssl_context'] = tls.get_ssl_context_from_settings(settings)
            
    try:
        prepare_cmd = args.prepare_command
    except AttributeError:
        # No sub-command specified, default to run command with stdin
        args.command = None
        args.script_hash = None
        args.signature = None
        args.resource = settings['stdin']
        prepare_cmd = run.prepare
    prepare_cmd(args, settings, parser)

    # Encrypt the resource if required:
    if settings['encryption_type'] in ('user_passphrase', 'random_passphrase'):
        settings['resource'] = gpg.encrypt_resource_symmetric(settings['resource'], settings['passphrase'])
    elif settings['encryption_type'] == 'recipient':
        settings['resource'] = gpg.encrypt_resource_to_recipients(settings['resource'], args.encrypt_to)                  
    
    return settings



import threading
import subprocess
import time
import logging

log = logging.getLogger('curlbomb.ssh')


class SSHRemoteForward(threading.Thread):
    def __init__(self, host, remote_forward, ssh_port=22):
        """Start an SSH connection to the specified host and remotely forward a port"""
        self.host = host
        self.ssh_port = str(ssh_port)
        self.remote_forward = remote_forward
        self._kill = False
        self._connected = False
        self._lines = []
        threading.Thread.__init__(self)

    def run(self):
        log.info("Creating ssh forward {} via {}".format(self.remote_forward, self.host))
        proc = subprocess.Popen(
            ['ssh','-v','-p',self.ssh_port,'-N','-R',self.remote_forward,self.host],
            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        while True:
            line = proc.stdout.readline()
            self._lines.append(line)
            if line == b'':
                self._kill = True
                break
            elif b'remote port forwarding failed' in line:
                self._kill = True
            elif b"All remote forwarding requests processed" in line:
                self._connected = True
                break
        while True:
            if self._kill:
                self.last_msg = self._lines[-2].decode("utf-8")
                proc.kill()
                break
            time.sleep(0.1)

    def wait_connected(self):
        try:
            while not self._kill:
                if self._connected:
                    log.info("SSH forward established")
                    return True
                time.sleep(0.1)
            return False
        except KeyboardInterrupt:
            self.kill()
    
    def kill(self):
        self._kill = True
        self.join()
        log.info("SSH connection closed: {}".format(self.host))



#!/usr/bin/env python

from setuptools import setup
import os

long_description = description = "A personal HTTP server for serving one-time-use shell scripts"
if os.path.exists('README.txt'):
    with open('README.txt') as f:
        long_description=f.read()

setup(name='curlbomb',
      version='1.4.1',
      description=description,
      long_description=long_description,
      license="MIT",
      classifiers=[
          "Development Status :: 4 - Beta",
          "Topic :: Utilities",
          "License :: OSI Approved :: MIT License",
      ],
      author='Ryan McGuire',
      author_email='ryan@enigmacurry.com',
      url='https://github.com/EnigmaCurry/curlbomb',
      packages=['curlbomb'],
      install_requires=['tornado','notify2', 'requests', 'psutil'],
      include_package_data = True,
      entry_points={
          'console_scripts': ['curlbomb = curlbomb:main']},
)

